/* ***** BEGIN LICENSE BLOCK *****
 * Licensed under Version: MPL 1.1/GPL 2.0/LGPL 2.1
 * Full Terms at http://mozile.mozdev.org/license.html
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Playsophy code (www.playsophy.com).
 *
 * The Initial Developer of the Original Code is Playsophy
 * Portions created by the Initial Developer are Copyright (C) 2002-2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * Richard Prescott
 *
 * ***** END LICENSE BLOCK ***** */

/**
@fileoverview 
<p>Mozilla Inline text editing that relies on eDOM, extensions to the standard w3c DOM.

<p>This file implements user modify in Mozilla by leverging Mozilla's proprietary
Selection object as well as eDOM, a set of browser independent extensions
to the w3c DOM for building editors. 

<p>2005-09 : Removed IE contentEditable references, no longer supported.
<p>Revised on 2004-04-18 : Added some namespace support

<hr />

<p>POST05:
<ul>
<li>refactor userModify code as part of "EditableElement"
<li>IE's "ContentEditable" means that an element is editable whenever its "ContentEditable"
setting is true. However, we may change this so that you have to set editing on for a
document as a whole before its individual editable sections become editable. This would
allow a user to browse an editable document and explicitly choose to edit it or not.
<li>see if can move to using DOM events and away from Window.getSelection() if possible 
(effects how generic it can be!)
<li>selection model: word, line etc. Write custom handlers of clicks and use new Range
expansion methods
</ul>

<p>POST04:
<ul>
<li>support for tracking whether changes were made to elements or not ie/ does a user
need to save? Should MozCE warn a user to save before exiting the browser? Some of
this may go into eDOM itself in enhancements to document or to all elements ie/ changed?
</ul>

<h4>References</h4>
<p>http://www.w3.org/TR/1999/WD-css3-userint-19990916#user-modify<br />
(Now outdated, see revised version.)
*/

/* === Mozilla Specific ===================================================== */
/**
mozCE: Gets mozUserModify (-moz-user-modify) computed style for this element.

<p>Start of "EditableElement": this will move into eDOM once it is fleshed out.

<p>POST04: set user-input and user-select properly as a side effect of setting user-modify.<br />
Need to chase to explicit parent of the editable area and check if true<br />
Also for contentEditable - make sure set moz user modify and other properties!

@return Mozilla proprietry read/write attribute
@type string
*/
Element.prototype.__defineGetter__(
	"mozUserModify",
	function() {return document.defaultView.getComputedStyle(this, null).MozUserModify}
);

/**
mozCE: Is this Element modifiable according to the MozUserModify computed style?.

@return true if set to "read-write"
@type boolean 
*/
Element.prototype.__defineGetter__(
	"mozUserModifiable",
	function() {
		if(this.mozUserModify=="read-write") return true;
		return false;
	}
);

/* === General ============================================================== */
/**
mozCE: Gets the userModify value for an element.

<p>Currently only tests the MozUserModify style but can be extended to include
other attributes as required.
 
@return "read-write", etc
@type string      
*/
Element.prototype.__defineGetter__(
	"userModify",
	function() {return this.mozUserModify}
);

/**
mozCE: Is this element modifiable?
 
@return true if "read-write", false otherwise.
@type boolean
*/
Element.prototype.__defineGetter__(
		"userModifiable",
		function(){
			if(this.userModify=="read-write") return true;
			return false;
		}
);

/**
mozCE: Gets the parent node (or this node) that is explicitly set to be userModifiable.

<p>Note that this accounts for different degrees of userModify.
If say "writetext" is inside a "write" then context will stop at the writetext
element. That is the context for that level of usermodify. 

@return node or null
@type Node
*/ 
Element.prototype.__defineGetter__(
	"userModifiableContext",
	function(){
		if(this.mozUserModifiable){
			var x = this;
			var um = this.mozUserModify;
			var pum;
			while(x.parentNode)	{
				pum = x.parentNode.mozUserModify;
				if(pum!=um) break;
				x = x.parentNode;
				um = pum;
			}
			return x;
		}
		return null;
	}
);

/* ===== SELECTION METHODS ================================================== */
/*
 Selection methods to support styling the current selection.
 Ideally there would be a new object called EditableSelection that specializes
 Selection and is accessible from Window.

 FUTURE:
 - move away from marked CSSRange once eDOM supports it.
 - redo as document.getEditableSelection() and a new object, EditableSelection.
*/

/**
mozCE: Selection Insertion Point.
Adds an insertion point Object (ip) to this Selection Object.

<p>POST05
<ul>
<li>May redefine SelectedInsertionPoint as a specialization of InsertionPoint.
 It can track changes in Selection.
<li>Maintain the insertion point - make it current.
</ul>

@param {boolean} collapse true collapses a selection to the start, false to the end.
@return ip: the insertion point
@type InsertionPoint
*/
Selection.prototype.insertionPoint = function(collapse)
{
	if(!this.isCollapsed)	{
		this.ip = null;
		if(collapse) { 
			this.collapseToStart();
		} else {
			this.collapseToEnd();
		}
	}

	var selectedip = null;
	var umc = this.anchorNode.parentElement.userModifiableContext;
	if(umc)
		selectedip = document.createInsertionPoint(umc, this.anchorNode, this.anchorOffset);
// if no current ip or if selected ip differs from current ip then reset current ip
//	if( !("ip" in this) || !this.ip || !selectedip || !selectedip.equivalent(this.ip)) {
	this.ip = selectedip;
	return this.ip;
}
/**
0.
@final
*/
Selection.MOVE_FORWARD_ONE = 0;
/**
1.
@final
*/
Selection.MOVE_BACK_ONE = 1;
/**
2.
@final
*/
Selection.MOVE_UP_ONE = 2;
/**
3.
@final
*/
Selection.MOVE_DOWN_ONE = 3;
/**
4.
@final
*/
Selection.MOVE_START_TOP = 4;
/**
5.
@final
*/
Selection.MOVE_END_TOP = 5;
/**
6.
@final
*/
Selection.MOVE_START_LINE = 6;
/**
7.
@final
*/
Selection.MOVE_END_LINE = 7;
/**
8.
@final
*/
Selection.DELETE_FORWARD = 8;
/**
9.
@final
*/
Selection.DELETE_BACKWARD = 9;

/**
mozCE: Move the caret relative to the current selection.
If the current Selection is a range rather than a point, the movement is relative
to the start point of the range. On exit the current selection is always collapsed
to the point that has been moved to.

@param {number} moveOption
@return true if IP moved.
@type boolean
*/
Selection.prototype.moveInsertionPoint = function(moveOption)
{
	var ip = this.insertionPoint(true);

	if(!ip)
		return false;

	switch(moveOption)
	{
		case Selection.MOVE_FORWARD_ONE:
			ip.forwardOne();
			break;
		case Selection.MOVE_BACK_ONE:
			ip.backOne();
			break;
		case Selection.MOVE_UP_ONE:
			ip.upOne();
			break;
		case Selection.MOVE_DOWN_ONE:
			ip.downOne();
			break;
		case Selection.MOVE_START_LINE:
			ip.toLineStart();
			break;
		case Selection.MOVE_END_LINE:
			ip.toLineEnd();
			break;
		case Selection.MOVE_START_TOP:
			ip.toTopStart();
			break;
		case Selection.MOVE_END_TOP:
			ip.toTopEnd();
			break;
		default: // no op for now - may become exception
			return false;
	}

	this.collapse(ip.ipNode, ip.ipOffset);
	
	return true;		
}

/**
mozCE: If this selection is collapsed and has CSS whitespace "pre" a tab character is inserted.

FUTURE:
Separate command to switch to the next editable area, with direction parameter.
E.g. for non-pre editing you could map tab for next area, ctrl+tab for previous area.

@return true if tab character inserted, false otherwise.
@type Boolean.
*/
Selection.prototype.insertTab = function( shift ) {
	if( this.isCollapsed ) {
		var ip = this.insertionPoint(true);
		if ( ip && ip.isPre() ) {
			ip.insertCharacter( 9 );
			this.collapse(ip.ipNode, ip.ipOffset);
			return true;
		}
	}
	return false;
}



/**
mozCE: Inserts a character at the current selection point.
If a range is selected the selected range is deleted and the character then inserted. 

@param {Integer} charCode
@return true if character inserted.
*/
Selection.prototype.insertCharacter = function(charCode)
{
	var cssr = this.getEditableRange(false);
	if(!cssr)
		return false;

	if(!cssr.collapsed) {
	// If there's a selection delete it.
		cssr.extractContentsByCSS();
	}
	var ip = document.createInsertionPoint(cssr.top, cssr.startContainer, cssr.startOffset);
	ip.insertCharacter(charCode);
	this.collapse(ip.ipNode, ip.ipOffset);
	return true;
}

/**
mozCE: XHTML Delete.

<p>"Delete" for selected XHTML represents three behaviors:

<ul>
<li>if range isn't collapsed then delete contents of the range - treat table contents properly (see code for behavior)
<li>if range is collapsed:
 <ul>
 <li>If at start of list line then outdent list line.
 <li>EIf at start of line then merge line with previous line, if there is one and this is appropriate.
 <li>Otherwise delete character or element before the selected point in the line
 </ul>
</ul>

<p>To do: Use cssr so that Bounded Line deletions can restore IP.

@param {number} direction
@return 1 if char deleted, 2 if structure deleted, 0 if nothing deleted.
@type Integer
*/
Selection.prototype.deleteSelection = function( direction )
{
	var ip, line, cssr, result;
	if(this.isCollapsed) {
		ip = this.insertionPoint(true);
		if(!ip)
			return 0;

		if (direction && direction==Selection.DELETE_FORWARD) {
			if ( ip.forwardOne() == InsertionPoint.AT_TOP ) {
				return 0;
			}
		}
		var deleted = ip.deletePreviousInLine();
		if(!deleted)	{
			line = ip.line;
			deleted = line.deleteStructure();
			if (deleted === null) {
			//At start of first list item.
				//(This is a hack because eDOM can't handle XHTML lists. Ideally should modify deleteStructure to give better return values.)
				var range = document.createRange();
				range.setStart( ip.ipNode, ip.ipOffset );
				range.collapse(true);
				cssr = document.createCSSTextRange( range, ip.top );
				eDOMXHTML.outdentLists(cssr);
				if (direction && direction==Selection.DELETE_FORWARD) {
				//If delete forward, step back one place.
				//Must do this because outdenting leaves the line with a container. (Could delete line container instead.)
					cssr.__restoreTextBoundaries();
					ip = cssr.firstInsertionPoint();
					ip.backOne();
					range = document.createRange();
					range.setStart( ip.ipNode, ip.ipOffset );
					range.collapse(true);
					cssr = document.createCSSTextRange( range, ip.top );
				}
				this.selectEditableRange(cssr);
				ip=this.insertionPoint();
			} else {
			//Line structure deleted ok.
				ip = deleted;
				//Problem if Bounded line. Need to restore text boundaries again???
				this.collapse(ip.ipNode, ip.ipOffset);
			}
			result=2;
		}
		else { //Simple char deletion, not start of line.
			this.collapse(ip.ipNode, ip.ipOffset);
			result=1;
		}
	}
	else { //range selected.
		cssr = this.getEditableRange(false);
		if(!cssr) {
			return 0;
		} else {
			cssr.extractContentsByCSS(); // delete selection
			this.removeAllRanges();
			this.addRange(cssr.cloneRange());
			this.collapseToStart();
			ip = this.insertionPoint();
			result=2;
		}
	}
	if (result>1) {
		line = ip.line;
		line.lineParent.__normalizeXHTMLTextStyle();
	}
	return result;
}

/**
mozCE: Splits an XHTML line or inserts a line break ("br").
If the current selection is not collapsed it is collapsed to the start and this is
used as the splitting point.

@param {boolean} inlineBreak true for an inline "br", false to split the line.
@return true if line split, otherwise false.
@type boolean
*/
Selection.prototype.splitXHTMLLine = function(inlineBreak)
{
	var cssr = this.getEditableRange(false);
	if(cssr) {
		if(!cssr.collapsed)	
			cssr.collapse(true);
		try {
			var ip = cssr.firstInsertionPoint();
			if (inlineBreak) {
				ip.breakXHTMLLine();
			}
			else {
				ip.splitXHTMLLine();
			}
			cssr.selectInsertionPoint(ip);
			this.selectEditableRange(cssr);
			return true;
		}catch(e){
			alert("Selection.splitXHTMLLine: "+e)
		}
	}
	return false;
}

/*
MAX - To Do. Maybe a special command just for XHTML class?
It would work similarly to the style commands.
If selection, wrap in a span and add class.
If no selection add class to line container.
Allow multiple classes.
If class already present remove it.
Would also need to normalize during delete operations, etc.
*/

/**
mozCE: Toggle Attribute.
Toggles a given attribute with a value or a default.
If the default is an empty string the attribute is removed if already present.
For text nodes the attribute is toggled on the parent of the text node.
There is not much control with this method, so use with care. 

@param {String} attribute
@param {String} attrValue
@param {string} attrDefault
*/
Selection.prototype.toggleAttribute = function(attribute, attrValue, attrDefault) {
	// alert("Anything");
	var contextNode;
	if(this.anchorNode.nodeType==Node.ELEMENT_NODE) {
		contextNode=this.anchorNode;
	}	else {
		contextNode=this.anchorNode.parentNode;
	}
	var current = contextNode.getAttribute(attribute);
	if(current==attrValue) {
		if(attrDefault=="") {
		// if the default attribute value is an empty string, then remove the attribute
			contextNode.removeAttribute(attribute);
		}	else {
		// otherwise set it to the default
			contextNode.setAttribute(attribute, attrDefault);
		}
	} else {
		contextNode.setAttribute(attribute, attrValue);
	}
}

/**
mozCE: Toggles CSS Style for this (text) selection.
Can handle multiple styles.
Error handling is limited, so make sure you get the MES commands correct.

POST05:
Change so defaultValue doesn't have to be passed in (i.e. remove style).
Think about toggling whole line if selection collapsed.

@param {Array} styleName : CSS style names (as Strings).
@param {Array} styleValue : CSS style values (as Strings) when toggling "on".
@param {Array} defaultValue : CSS style values (as Strings) when toggling "off".
*/
Selection.prototype.toggleTextStyle = function(styleName, styleValue, defaultValue)
{
	var cssr = this.getEditableRange(true);
	if(cssr) {
		try {
			if (!styleName[0] || !styleValue[0] || !defaultValue[0] ) throw "missing value";
			var value = (cssr.hasStyle(styleName[0], styleValue[0])) ? defaultValue : styleValue;
			for ( var s=0; s<styleName.length; s++ ) {	
				if (!styleName[s] || !value[0] ) throw "missing value";
				cssr.styleText(styleName[s], value[s]);
			}
		}catch(e){
		}
		this.selectEditableRange(cssr);
	}
}

/**
mozCE: Applies CSS Style to this (text) selection.
POST05: Think about toggling whole line if selection collapsed.

@param {String} styleName CSS style name.
@param {String} syleValue CSS style value.
*/
Selection.prototype.styleText = function(styleName, styleValue)
{
	var cssr = this.getEditableRange(true);
	if(cssr) {
		cssr.styleText(styleName, styleValue);
		this.selectEditableRange(cssr);
	}
}

/**
mozCE: Wraps this selection with element node "elementName".

@param {String} namespace
@param {String} elementName
*/
Selection.prototype.wrapTextNS = function(namespace, elementName) {
	try {
		var cssr = this.getEditableRange(true);
		if(cssr) {
			cssr.wrapTextNS(namespace, elementName);
			this.selectEditableRange(cssr);
		}
	} catch(e) {
		alert("wrapText: "+e);
	}
}

/**
mozCE: Adds an XHTML text link to this selection, an existing link can be edited.

<p>A dialog is displayed in which the link reference ("href" attribute) is entered.

<p>The first range in the selection is enclosed by the XHTML link tag &lt;a&gt;.
The selected range remains selected on exit.

<p>If a null string is entered, "clearTextLinks" is executed instead.

<p>If the range is collapsed an XHTML link ancestor is searched for and selected if found.
The current "href" value is extracted and shown in the text entry box of the dialog
where it can be edited. The cursor position is restored after addding the link.
*/
Selection.prototype.linkText = function()
{
	var reCollapse = false;
	var cssr = this.getEditableRange(true);
	if (cssr) {
		var href="";

		if(cssr.collapsed) {
			var a = cssr.startContainer.getLinkAncestor();
			if (a) {
				//Now convert to CSS Text Range. 
				var firstIP = a.firstInsertionPoint(cssr.top);
				var lastIP = a.lastInsertionPoint(cssr.top);
				var top = cssr.top;
				var cssr2 = __createCSSRangeFromIPs(top, firstIP, lastIP);
				cssr.detach();
				cssr = cssr2;
				reCollapse = true;
				if (a.hasAttribute("href")) {
					href = a.getAttribute("href");
				}		
			}
			else {
				//Not a hyperlink,do nothing.
				return;
			}
		} // else selection ... must not already contain an <a>?

		if(!cssr.collapsed) {
			href = prompt("Enter a URL:", href);
			if(href == null) {
				// Prompt cancelled
			}else if(href != "") {
				cssr.linkText(href);
				if (!reCollapse)
					this.selectEditableRange(cssr);
				//else
					//Selection won't have changed so no need to restore.
			}else {
				this.clearTextLinks();
			}
		}
	}				
}


/*
mozCE: Clear XHTML text links from this selection.

<p>If the first range in the selection is editable all links are removed.
The selected range remains selected on exit.

<p>If the range is collapsed an XHTML link ancestor is searched for and selected if found.
The cursor position is restored after removing the link.
*/
Selection.prototype.clearTextLinks = function()
{
	try {
		var reCollapse = false;
		var cssr = this.getEditableRange(true);
		if (cssr) {
			if(cssr.collapsed) {
				var a = cssr.startContainer.getLinkAncestor();
				if (a) {
					//Save current absolute text position.
					var textPoint = new __TextPointer( a.parentNode, cssr.startContainer, cssr.startOffset );
					//Now convert to CSS Text Range. 
					var firstIP = a.firstInsertionPoint(cssr.top);
					var lastIP = a.lastInsertionPoint(cssr.top);
					var top = cssr.top;
					var cssr2 = __createCSSRangeFromIPs(top, firstIP, lastIP);
					cssr.detach();
					cssr = cssr2;
					reCollapse=true;
				} else {
					//Not a hyperlink,do nothing.
					return;
				}
			}

			if (cssr) {
				cssr.clearTextLinks();
				if (reCollapse) {
					//Restore cursor position.
					var refTextNode = textPoint.referencedTextNode();
					var refOffset = textPoint.referencedOffset();
					this.collapse( refTextNode, refOffset);
					delete textPoint;
				} else {
					this.selectEditableRange(cssr);
				}
			}
		}
	}catch(e){
	}
}

/**
mozCE: Applies CSS Style to Lines.
i.e. Applies the style to the containers of the Contained Lines in the editable
part of this selection.
Bounded lines are converted to Contained Lines using the
<code>defaultContainer</code> as the container.
A container is also added if the container is top.

@param {String} styleName CSS property-name
@param {String} styleValue CSS value
*/
Selection.prototype.styleLines = function(styleName, styleValue)
{
	var cssr = this.getEditableRange(true);

	if(!cssr)
		return;

	var lines = cssr.lines;	

	for(var i=0; i<lines.length; i++) {
		var thisLine = lines[i];
		if ( ((thisLine.lineType == CSSLine.BOUNDED_LINE) && !(thisLine.emptyLine))
		 ||  ((thisLine.lineType == CSSLine.CONTAINED_LINE) && (thisLine.container == thisLine.top)) ) {
			var container = eDOMXHTML.createElement(eDOMXHTML.defaultContainer);
			if ( thisLine.lineParent.canAppendChild( container ) !== false ) {
				thisLine = thisLine.setContainer(container, false);
				thisLine.normalizeWhitespace();
			}
		}
		//If Contained Line (or Bounded Line that has been converted to a Contained Line above.)
		if ( thisLine.lineType == CSSLine.CONTAINED_LINE ) {
			if ( thisLine.container.descendent( thisLine.top ) ) {
			//If line is a descendent of "top";
				thisLine.setStyle(styleName, styleValue);
				var lp = thisLine.lineParent;
				lp.__normalizeXHTMLTextStyle();
				lp.removeRedundantStyles();
			}
		}
	}

	this.selectEditableRange(cssr);
	cssr.detach();
}

/**
mozCE: XHTML Change Lines Container

<p>Note: Don't use this format command unless you are formatting lines because
it always creates the containerName for the top line if there isn't a container.
i.e. it doesn't check that containerName is a valid descendent of the container,
it changes/inserts it regardless.
So you could end up with a &lt;p&gt; element embedded within a &lt;p&gt; element.
For example if you tried something like:<br />
&lt;p class="editable"&gt;Edit this paragraph text&lt;/p&gt;

<p>You would end up with:<br />
&lt;p class="editable"&gt;&lt;p&gt;Edit this paragraph text&lt;/p&gt;&lt;/p&gt;

<p>This is only a potential problem if you have multiple editable areas because you can
only specify one MES file. So if you have editable div's and editable p's you
can end up running into problems.

<p>Possibly this function should check that containerName is a valid descendent of its line parent
before doing anything?

<p>POST05: XHTML extensions.
- For "br" within a Bounded Line don't apply container directly.
Promote the "br" to a line (with an editing token) and delete the "br".
(MAX: If this is done I think it should be an XHTML specific command in addition to the standard
changeLinesContainerNS which should be generic.) 
- Promotion can also mean moving a line outside of its parent.(table el => make div and promote 
outside of it)
*/
Selection.prototype.changeLinesContainerXHTML = function(containerName)
{
	this.changeLinesContainerNS(eDOMXHTML.NS,containerName);
}
/**
mozCE: Change Lines Container.
Replaces the container element for all the Contained Lines in the selected range.
If this is a Bounded Line <code>containerName</code> is created - the Bounded Line becomes
a Contained Line.

@param {String} namespace optional.
@param containerName new container name.
*/
Selection.prototype.changeLinesContainerNS = function(namespace, containerName)
{
	var cssr = this.getEditableRange(true);
	if(!cssr)
		return;

	var replaceContainer, container, clType;
	var lines = cssr.lines;

	//Do all Bounded Lines first.
	//Must do this, otherwise start boundaries disappear if the container of the previous line is changed.
	for(var i=0; i<lines.length; i++) { 
		replaceContainer = null;

		if (!namespace) {
			container = document.createElement(containerName);
		} else {
			container = document.createElementNS(namespace, containerName);
		}

		if ( lines[i].lineType == CSSLine.BOUNDED_LINE ) {
			// container name must be a valid descendent.
				if (lines[i].lineParent.canAppendChild(container)) {
					replaceContainer = true;
				}
		}
		if ( replaceContainer !== null ) {
			lines[i] = lines[i].setContainer(container, replaceContainer);
			lines[i].normalizeWhitespace();
		}
	}

	//Now do all Contained Lines.
	for(var i=0; i<lines.length; i++) { 
		replaceContainer = null;

		if (!namespace) {
			container = document.createElement(containerName);
		} else {
			container = document.createElementNS(namespace, containerName);
		}

		if ( lines[i].lineType == CSSLine.CONTAINED_LINE ) {
			clType = lines[i].containedLineType;
			if ( lines[i].container.nodeName == container.nodeName ) {
				//No need to replace matching container. 
			} else if ( cssr.top.descendent(lines[i].container) ) {
				//do nothing, line container is out of the editable range.
			} else if ( clType == ContainedLine.TOP ) {
				if (lines[i].container.canAppendChild(container)) {
					replaceContainer = false;
				}
			} else if ( clType != ContainedLine.BLOCK ) {
				// keep container if it is a table cell or list item.
				if (lines[i].container.canAppendChild(container)) {
					replaceContainer = false;
				}
			}	else if ( clType == ContainedLine.BLOCK ) {
				if (lines[i].container.parentNode.canAppendChild(container)) {
					replaceContainer = true;
				}
			}
		}
		if ( replaceContainer !== null ) {
			lines[i] = lines[i].setContainer(container, replaceContainer);
			lines[i].normalizeWhitespace();
		}
	}

	this.selectEditableRange(cssr);
	cssr.detach();
}


/**
mozCE: Removes Lines Container.
Removes the container for all the Contained Lines in the selected range.
Has no effect on Bounded Lines.
*/
Selection.prototype.removeLinesContainer = function()
{
	var cssr = this.getEditableRange(true);
	if(!cssr) return;
	var lines = cssr.lines;
	for(var i=0; i<lines.length; i++) {
		if( lines[i].lineType == CSSLine.CONTAINED_LINE ) {
		//If Contained Line
			var clType = lines[i].containedLineType;
			if ( clType == ContainedLine.BLOCK ) {
			//If BLOCK (...not list or table cell)
				if (lines[i].container.descendent(cssr.top)) {
				//If container is within the editable area
					lines[i].removeContainer();
				}
			}
		}
	}
	this.selectEditableRange(cssr);
}
/**
mozCE: Indent Lines.
Optional <code>indent</code> : default is "+40px".

@param {String} indent CSS value.
*/
Selection.prototype.indentLines = function( indent )
{	
	if (!indent)
		indent = "+40px";
	var cssr = this.getEditableRange(true);
	if(cssr){
		eDOMXHTML.indentLines(cssr, indent);	
		this.selectEditableRange(cssr);
	}
}
/**
mozCE: Outdent Lines.
Optional <code>outdent</code> : default is "-40px".

@param {String} outdent CSS value.
*/
Selection.prototype.outdentLines = function( outdent )
{	
	if (!outdent)
		outdent = "-40px";
	this.indentLines( outdent );
}
/**
mozCE: Toggle List Lines.
*/
Selection.prototype.toggleListLines = function(listContainerName, listContainerToChange)
{	
	var cssr = this.getEditableRange(true);
	if(cssr){
		eDOMXHTML.listLinesToggle(cssr, listContainerName, listContainerToChange);
		this.selectEditableRange(cssr);
	}
}
/**
mozCE: indent Lists.
*/
Selection.prototype.indentLists = function()
{	
	var cssr = this.getEditableRange(true);
	if(cssr){
		eDOMXHTML.indentLists(cssr);
		this.selectEditableRange(cssr);
	}
}
/**
mozCE: outdent Lists
*/ 
Selection.prototype.outdentLists = function()
{	
	var cssr = this.getEditableRange(true);
	if(cssr){
		eDOMXHTML.outdentLists(cssr);
		this.selectEditableRange(cssr);
	}
}
/**
mozCE: Inserts <code>node</code> at the first insertion point for this selection.

To Do:
This seems rather dangerous because some insert operations may
require normalizations to be performed. (e.g. from a Paste operation.)
I've added a test for text nodes, but if you are adding element nodes you
will have to call the normalization routines separately.
Maybe for XHTML have a separate insertXHTMLNode function that then
calls the required XHTML tidy up functions?

@param {Node} node
*/
Selection.prototype.insertNode = function(node)
{
	var cssr = this.getEditableRange(true);
	if(!cssr)
		return false;
	//
	//Create insertion point for first ip in this range.
	var ip = cssr.firstInsertionPoint();

	//Insert new node.
	ip.insertNode(node);

	//Collapse selection to start of inserted node.
	cssr.__clearTextBoundaries(); // POST05: don't want to have to use this
	cssr.selectInsertionPoint(ip);

	//If text node inserted, normalize text nodes.
	// ... always normalize to be on the safe side???
//	if ( node.nodeType == Node.TEXT_NODE )
		cssr.normalizeText();

	this.selectEditableRange(cssr);
	return true;
}
/**
Insert XHTML element.

@param {String} name
@return false if error
@type boolean
*/
Selection.prototype.insertXHTMLElement = function( name )
{
	var element = eDOMXHTML.createElement(name);
	if (element)
		return this.insertNode(element);
	return false;
}

Selection.prototype.insertXHTMLImg = function()
{
	var imgref = prompt("Enter the image url or file name:", "");
	if(imgref == null) // null href means prompt canceled
		return;
	if(imgref == "")
		return; // ok with no name filled in
	var img = eDOMXHTML.createElement("img");
	img.src = imgref; // any way to tell if it is valid?
	this.insertNode(img);
}
/**
mozCE: Get editable range of this selection.

<p>Shorthand way to get CSS Range for the current selection.
This range will be marked; hence it can easily be restored.

<p>Only uses the first range in the selection.

<p>POST05: 
<ul>
<li>consider not calculating textpointers here (createCSSTextRange) but only within the
  editing functions in eDOM where the context can be given more precisely.
<li>allow text selection to only begin and end on word boundaries (part of CSSTextRange 
selection methods)
<li>account for empty editable area (maybe in isContentEditable); account for selection
type ie/ element or object or text.
</ul>

@param {boolean} mark: optional, "true" marks the range for later reference.
@return CSS Range or null
@type Range
*/
Selection.prototype.getEditableRange = function(mark)
{	
	try	{
		var range = this.getRangeAt(0);
		if(!range.commonAncestorContainer.parentElement.userModifiable)
			return null;
		range = range.cloneRange();
		var top = range.commonAncestorContainer.parentElement.userModifiableContext;
		range = __createCSSRange(range, top, mark);
		return range;
	}
	catch(e) {
		return null;
	}
}

/*
mozCE: Select editable range. 
Restore the range.
*/
Selection.prototype.selectEditableRange = function(cssr)
{
	try {
		cssr.__restoreTextBoundaries();
		this.removeAllRanges();
		this.addRange(cssr.cloneRange());
	}
	catch(e)
	{}	
}